/*Popa Emilia Ioana*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/sleep.h>
#include <avr/wdt.h>
#include <string.h>
#include <stdio.h>
#include "diskio.h"
#include "pff.h"
#include "lcd.h"

#define F_CPU 16000000

#define PAUSE_MAX_DISTANCE	30

/* Four-Character Code
 *
 * Folosit pentru a indentifica formate de date
 */
#define FCC(c1, c2, c3, c4) \
	(((DWORD)c4 << 24) + ((DWORD)c3 << 16) + ((WORD)c2 << 8) + (BYTE)c1)

unsigned char Buff1[256];	// Wave output FIFO
unsigned char Buff2[256];	// Wave output FIFO 
volatile uint8_t read = 1;
volatile uint16_t counter = 0;
volatile uint8_t sync = 0;
FATFS Fs;	// sistemul de fisiere
DIR Dir;	// un director
FILINFO music_file1;
FILINFO music_file2;

FILINFO music_files[10];
int music_files_no = 0;
int current_music_file = -1;

uint16_t miliseconds = 0;
uint8_t minutes = 0, seconds = 0, hours = 0;
char time_string[20];

char states[3][10] = { "STOP", "PLAY", "PAUSE" };

volatile unsigned long int distance = 0;
volatile int dd = 0;
volatile unsigned long int microseconds = 0;

#define NONE -1
#define BTN_PB0 1
#define BTN_PB1 2
#define STOP	0
#define PLAY	1
#define PAUSE	2
#define NEXT	3

//controleaza redarea audio
volatile uint8_t continue_play = PLAY;
//memoreaza ce buton a fost apasat
volatile uint8_t button_pressed = NONE;

void setLEDS(int);

int DIST_get();

ISR(TIMER0_COMPA_vect)
{
	if( continue_play == PLAY) {
		if(sync)
		{
			if(read)
			{
				OCR1A = (unsigned char) Buff1[counter++];

				if(counter == 256)
				{
					read = 0;
					counter = 0;
				}
			} else {
				OCR1A = (unsigned char)Buff2[counter++];

				if(counter == 256)
				{
					read = 1;
					counter = 0;
				}
			}

			if(OCR1A > 150) {
				setLEDS(3);
			} else {
				setLEDS(0);
			}
		}
	}
}

ISR(TIMER2_COMPA_vect)
{
	microseconds ++;
}

void Timer0_init(void)
{
	OCR0A = 0x80;
	TCCR0A = (1 << WGM01);	// mod CTC top OCR0A
	TCCR0B = (1 << CS01);	// prescaler 8
	TIMSK0 = (1 << OCIE0A);	// intrerupere de output compare A
}

void Timer1_init(void)
{
	OCR1A = 0xFF;
	TCCR1A = (1 << COM1A1) | (1 << COM1A0) | (1 << WGM10);	// mod fastPWM 8-bit; set OC0A on match, clear on bottom
	TCCR1B = (1 << WGM12) | (1 << CS10); 			// prescaler 8
}

void Timer2_init(void)
{
	OCR2A = 0x10;
	TCCR2A = (1 << WGM21);			// mod CTC
	TCCR2B = (1 << CS21) | (1 << CS20);			// prescaler 1
	TIMSK2 = (1 << OCIE2A);			// intrerupere de compare pe A 
}

void IO_init(void)
{
	DDRD |= (1 << PD5);			// OC1A ca output

	// cele 2 butoane pentru selectarea melodiei urmatoare sau precedente
	DDRB &= ~((1 << PB1) | (1 << PB0));	// PB0, PB1 ca input
	PORTB |= (1 << PB1) | (1 << PB0);	// pull-up-uri pentru PB0 si PB1

	// pini folositi pentru modificarea starii celor 6 led-uri
	DDRD |= (1 << PD0);   // pinul corespunzator led-ului USER este output
	DDRD |= (1 << PD1);   // pinul corespunzator led-ului USER este output
	DDRD |= (1 << PD2);   // pinul corespunzator led-ului USER este output

}

void setLEDS(int level) {
	PORTD &= ~((1 << PD0) | (1 << PD1) | (1 << PD2));
	if(level == 1) {
		PORTD |= (1 << PD0);
	}
	if(level == 2) {
		PORTD |= (1 << PD0) | (1 << PD1);
	}
	if(level == 3 ) {
		PORTD |= (1 << PD0) | (1 << PD1) | (1 << PD2);
	}
}

//WORD ret;
DWORD load_header(void)
{
	DWORD size;
	WORD ret;

	/* Citeste header-ul (12 octeti) */
	if(pf_read(Buff1, 12, &ret))
		return 1;

	if(ret != 12 || LD_DWORD(Buff1 + 8) != FCC('W','A','V','E'))
		return 0;

	for(;;)
	{
		/* Citeste Chunk ID si size */
		pf_read(Buff1, 8, &ret);
		if (ret != 8)
			return 0;

		size = LD_DWORD(&Buff1[4]);

		/* Verifica FCC */
		switch (LD_DWORD(&Buff1[0]))
		{
			/* 'fmt ' chunk */
			case FCC('f','m','t',' '):
				if (size > 100 || size < 16) return 0;			// Verifica size

				/* Citeste continutul */
				pf_read(Buff1, size, &ret);
				if (ret != size || Buff1[0] != 1) return 0;		// Verifica codificarea
				if (Buff1[2] != 1 && Buff1[2] != 2) return 0;		// Verifica canalele
				if (Buff1[14] != 8 && Buff1[14] != 16) return 0;	// Verifica resolutia

				/* Seteaza frecventa */
				OCR0A = (BYTE)(F_CPU / 8 / LD_WORD(&Buff1[4])) - 1;
				break;

			/* 'data' chunk => incepe redarea */
			case FCC('d','a','t','a') :
				return size;

			/* 'LIST' chunk => skip */
			case FCC('L','I','S','T') :
			/* 'fact' chunk => skip */
			case FCC('f','a','c','t') :				
				pf_lseek(Fs.fptr + size);
				break;

			/* Chunk necunoscut => eroare */
			default :					
				return 0;
		}
	}

	return 0;
}

void printState(int state) {
	LCD_clear_bottom_line();
	LCD_printAt(0x40,states[state]);
}

/*
 * Functie care reda un fisier audio
 *
 * path - calea absoluta a fisierului
 *
 * @return 0 daca a rulat cu succes fisierul
 */
UINT play (const char *path)
{
	WORD ret;
	FRESULT res;
	unsigned char read2;
	uint16_t buf_size = 256;

	DWORD current_size;
	unsigned long int n = 0;

	unsigned long int dist = 0;

	if((res = pf_open(path)) == FR_OK)
	{

		Timer0_init();
		Timer1_init();

		/* Incarca header-ul fisierului */
		current_size = load_header();		

		if(current_size < 256)
			return (UINT)current_size;

		/* Snip sector unaligned part */
		pf_read(Buff1, 512 - (Fs.fptr % 512), &ret);

		res = pf_read(Buff1, buf_size, &ret);
		if(res != FR_OK)
			return res;

		sync = 1;
		do
		{	
			n ++;
			if(n == 100) {
				dist = DIST_get();
				n = 0;

				if(dist < PAUSE_MAX_DISTANCE) {
					continue_play = PAUSE;
				}
			}

			if(continue_play == PAUSE) {
				n = 0;
				while(DIST_get() < PAUSE_MAX_DISTANCE) {
					printState(PAUSE);
					_delay_ms(100);
					n++;
					if(n == 10) {
						continue_play = NEXT;
					}
				}
				n = 0;

				if(continue_play != NEXT) {
					while(DIST_get() > PAUSE_MAX_DISTANCE) { _delay_ms(300); }
					continue_play = PLAY;
					printState(PLAY);
				}
			}

			if(continue_play != NEXT) {
				if((PINB & (1 << PB1)) == 0) {
					continue_play = NEXT;
					_delay_ms(100);
				}
			}

			if(continue_play != NEXT) {
				if((PINB & (1 << PB0)) == 0) {
					for(n = 0; n < 2; n++) {
						current_music_file --;
						if(current_music_file < 0) {
							current_music_file = music_files_no - 1;
						}
					}
					continue_play = NEXT;
					_delay_ms(100);
				}
			}

			read2 = read;
			if(continue_play == NEXT) {
				continue_play = PLAY;
				break;
			}

			if(read == 1)	
				res = pf_read(Buff2, buf_size, &ret);
			else
				res = pf_read(Buff1, buf_size, &ret);

			if(res != FR_OK)
				break;

			while(read2 == read);


		} while(continue_play == PLAY && ret == buf_size);	// Repeta pana cand toate datele au fost citite

		sync = 0;
		res = 0;
	}

	return res;
}

void getMusicFiles(const char *dir) {


	FILINFO info;
	//daca fisierele au fost citite intial nu mai citim
	if(music_files_no > 0) {
		return;
	}

	if(pf_opendir(&Dir, dir)) {
		LCD_print("ERROR OPENDIR");
		_delay_ms(2000);
		return;
	}

	while (!pf_readdir(&Dir, &info) && music_files_no < 6)
	{
		if(info.fname != NULL && info.fsize > 0 && info.fname[0] && (strstr(info.fname, ".WAV") || strstr(info.fname, ".wav"))) {

			if (!(info.fattrib & (AM_DIR|AM_HID)) ) {
				music_files[music_files_no] = info;
				music_files_no++;
			}
		}
	}
}

/*
 * Seteaza timpul curent
 *
 * @param h - ora curenta
 * @param m - minutul curent
 * @param s - secunda curenta
 * @param mili - milisecunda curenta
 */
void set_time(uint8_t h, uint8_t m, uint8_t s, uint16_t mili)
{
	int8_t activate = 0;

	if(TIMSK2 & (1 << OCIE2A)) {
		activate = 1;
		TIMSK2 &= ~(1 << OCIE2A);
	}

	hours = h;
	minutes = m;
	seconds = s;
	miliseconds = mili;

	if(activate)
		TIMSK2 = (1 << OCIE2A);
}

void start_play(){
	char time[20];
	char dir[10] = {"music"};

	getMusicFiles("music");

	if(music_files_no <= 0) return;
	printState(PLAY);
	current_music_file ++;
	if(current_music_file >= music_files_no) {
		current_music_file = 0;
	}

	FILINFO music_file = music_files[current_music_file];

	LCD_clear_top_line();
	LCD_printAt(0x00,music_file.fname);

	sprintf(time, "%s/%s",dir,music_file.fname);
	play(time);
	
}

ISR(INT2_vect) {
	if((PINB & (1 << PB2)) == 0 ) {
		dd = 2;
		distance = microseconds;
	} else {
		microseconds = 0;
		dd = 1;
	}
}

void DIST_init() {
	EIMSK |= (1 << INT2);
	EICRA |= (1 << ISC20);

	DDRD |= (1 << PD7);
	DDRB &= ~(1 << PB2);
	PORTB |= (1 << PB2);

}

int DIST_get() {

	dd = 0;
	distance = 0;
	PORTD &= ~(1 << PD7);
	_delay_us(2);
	PORTD |= (1 << PD7);
	_delay_us(10);
	PORTD &= ~(1 << PD7);

	while(dd != 2) {}

	return distance;
}

int main (void)
{

	DIST_init();
	sei();
	IO_init();
	LCD_init();
	Timer2_init();
	
	LCD_clear_top_line();
	LCD_printAt(0x00,"INITIALIZE");
	_delay_ms(2000);

	button_pressed = NONE;
	for(;;)
	{	/* Monteaza sistemul de fisiere */
		while(pf_mount(&Fs) != FR_OK)
		{
			LCD_printAt(0x00, "mounting...");

			// wait a while and retry
			_delay_ms(1000);
			
		}
		

		LCD_clear_top_line();
		LCD_printAt(0x00,"FS MOUNTED");
		_delay_ms(2000);
		while(1) {
			
			start_play();
			
			
		}

	}

	return 0;
}
